"""
PyHP: Python Hypertext Preprocessor
"""

from argparse import ArgumentParser
from pathlib import PurePath, Path

try:
    from pyhp_interface import Pyhp
    from file_processing import SystemFileProcessor
    from pyhp_flask import create_app
except ImportError:
    from .pyhp_interface import Pyhp
    from .file_processing import SystemFileProcessor
    from .pyhp_flask import create_app


if __name__ == '__main__':
    parser = ArgumentParser(description='Python Hypertext Preprocessor (PyHP)')
    parser.add_argument('-d', '--debug', action='store_true', help='Debug mode',
                        default=False)

    action_parser = parser.add_subparsers(dest='action',
                                          help='Action to perform',
                                          required=True)
    file_parser = action_parser.add_parser('file', help='Run a file')
    file_parser.add_argument('file', help='File to run')

    server_parser = action_parser.add_parser('server', help='Run a server')
    server_parser.add_argument('directory', help='Directory to serve')
    server_parser.add_argument('--port', help='Port to serve on', type=int,
                               default=5000)

    args = parser.parse_args()

    if args.action == 'file':
        base_dir = Path(args.file).parent.absolute()
        root_pyhp = Pyhp(PurePath(), SystemFileProcessor(base_dir),
                         args.debug)
        print(root_pyhp.run(PurePath(args.file).name))
    elif args.action == 'server':
        app = create_app(args.directory)
        app.run(port=args.port, debug=args.debug)
    else:
        parser.print_help()



"""
Tests PyHP separate from any server.

TestPyhpRemoveInitialIndentation:
    Tests that the initial indentation is correctly removed.
TestPyhpPrepareCodeText:
    Tests that the code text is correctly prepared
    (excess newlines removed, etc)
TestPyhpPrepareContext:
    Tests that the context (globals and locals) contains the required
    information.
TestPyhpRunParsedCode:
    Tests that the code is correctly run, including cookies, GET, POST.
TestPyhpFileProcessing:
    Tests that PyHP can load and execute files.
"""

# pylint: disable=missing-function-docstring

from unittest import TestCase
from datetime import datetime
from pathlib import PurePath

try:
    from mocks import MockFileProcessor
except ImportError:
    from .mocks import MockFileProcessor

from src.pyhp.text_processing import remove_initial_indentation, \
    prepare_code_text
from src.pyhp.hypertext_processing import parse_text
from src.pyhp.code_execution import run_parsed_code
from src.pyhp.cookies import NewCookie, DeleteCookie
from src.pyhp.pyhp_interface import Pyhp


class TestPyhpRemoveInitialIndentation(TestCase):
    """Tests that the initial indentation is correctly removed."""

    def test_no_indentation(self):
        cases = (
            '',
            "print('Hello')",
            '''if True:
    print('True')
else:
    print('False')''',
            'foo = [x**x for x in range(10)]',
            '''print('<pyhp>print("Hello")</pyhp>')''',
        )

        for string in cases:
            self.assertEqual(string, remove_initial_indentation(string))

    def test_indented(self):
        cases = (
            (' x = 1', 'x = 1'),
            ('    print(1)', 'print(1)'),
            ('\tprint(2)', 'print(2)'),
            ('\t  print(3)', 'print(3)'),
            ('\t    if True:\n\t        print(1)', 'if True:\n    print(1)'),
        )

        for indented, unindented in cases:
            self.assertEqual(unindented, remove_initial_indentation(indented))

    def test_less_indentation_in_subsequent_lines_than_first_line(self):
        error_cases = [
            '    print(1)\nprint(2)',
            '  x=1\n x=2',
            '\t\tprint(2)\n\t\tprint(3)\n\tprint(5)',
        ]
        no_error_cases = [
            '    if True:\n        print(1)\n\n        print(2)',
            ' print(1)\n\n\n print(2)',
        ]

        for case in error_cases:
            self.assertRaises(SyntaxError, remove_initial_indentation, case)
        for case in no_error_cases:
            remove_initial_indentation(case)

    def test_multiline_string_indentation(self):
        pass  # TODO: Finish


class TestPyhpPrepareCodeText(TestCase):
    """
    Tests that the code text is correctly prepared
    (excess newlines removed, etc)
    """

    def test_extra_newlines(self):
        cases = [
            ('x = 1\n\n', 'x = 1'),
            ('\n\nx = 1\n\n', 'x = 1'),
            ('\n\nx = 1\n\n\n', 'x = 1'),
            ('print("Hello\\nWorld")', 'print("Hello\\nWorld")'),
            ('print("Hello")\nprint("World")',
             'print("Hello")\nprint("World")'),
        ]

        for input_code, output in cases:
            self.assertEqual(prepare_code_text(input_code), output)


class TestPyhpPrepareContext(TestCase):
    """Tests that the globals and locals contain the required information."""

    def test_typical_globals(self):
        file_processor = MockFileProcessor()
        pyhp_class = Pyhp(PurePath(), file_processor)
        self.assertIs(pyhp_class.globals['pyhp'],
                      pyhp_class)


class TestPyhpRunParsedCode(TestCase):
    """Tests that the code is correctly run, including cookies, GET, POST."""

    def test_normal_html(self):
        cases = [
            '<p>Hello</p>',
            'Hi',
            '<div>Hello</div>',
        ]

        for case in cases:
            self.assertEqual(create_file_processor_and_run_code(case), case)

    def test_blank_pyhp_tags(self):
        cases = [
            '<pyhp></pyhp>',
            '<pyhp> </pyhp>',
            '<pyhp>\n</pyhp>',
            '<pyhp>\n\n</pyhp>',
            '<pyhp></pyhp><pyhp></pyhp>',
        ]

        for case in cases:
            self.assertEqual(create_file_processor_and_run_code(case), '')

    def test_print_statement(self):
        cases = [
            ("<pyhp>print('Hello', end='')</pyhp>", 'Hello'),
            ("<pyhp>print('Hello', end=' ')\nprint('World', end='')</pyhp>",
             'Hello World'),
        ]

        for input_code, expected in cases:
            self.assertEqual(create_file_processor_and_run_code(input_code),
                             expected)

    def test_get_cookies(self):
        pass  # TODO: Finish

    def test_set_cookies(self):
        cases = [
            ("<pyhp>pyhp.set_cookie('foo', value='bar')</pyhp>",
             {'foo': NewCookie('foo', 'bar')}),
            ("<pyhp>pyhp.set_cookie('foo', value='bar', max_age=10)</pyhp>",
             {'foo': NewCookie('foo', 'bar', max_age=10)}),
            ("<pyhp>from datetime import datetime, timedelta\n"
             "pyhp.set_cookie('foo', value='bar', "
             "max_age=10, expires=datetime(2030, 1, 1))</pyhp>",
             {'foo': NewCookie('foo', 'bar', max_age=10,
                               expires=datetime(2030, 1, 1))}),
        ]

        for case in cases:
            file_processor = MockFileProcessor()
            pyhp_class = Pyhp(PurePath(), file_processor)
            run_parsed_code(parse_text(case[0]), pyhp_class)
            self.assertEqual(case[1], pyhp_class.get_new_cookies())

    def test_delete_cookies(self):
        # Create a cookie then delete it
        file_processor = MockFileProcessor()
        pyhp_class = Pyhp(PurePath(), file_processor)
        code = '<pyhp>pyhp.set_cookie("foo", value="bar")\n' \
               'pyhp.delete_cookie("foo")</pyhp>'
        run_parsed_code(parse_text(code), pyhp_class)
        self.assertEqual({'foo': NewCookie('foo', 'bar')},
                         pyhp_class.get_new_cookies())
        self.assertEqual({'foo': DeleteCookie('foo')},
                         pyhp_class.get_delete_cookies())

        # Start with a cookie then delete it
        pyhp_class = Pyhp(PurePath(), file_processor, cookies={'foo': 'bar'})
        code = '<pyhp>pyhp.delete_cookie("foo")</pyhp>'
        run_parsed_code(parse_text(code), pyhp_class)
        self.assertEqual({'foo': DeleteCookie('foo')},
                         pyhp_class.get_delete_cookies())

        # Delete an unexisting cookie
        pyhp_class = Pyhp(PurePath(), file_processor)
        code = '<pyhp>pyhp.delete_cookie("foo")</pyhp>'
        run_parsed_code(parse_text(code), pyhp_class)
        # TODO: Should this be empty?
        self.assertEqual({'foo': DeleteCookie('foo')},
                         pyhp_class.get_delete_cookies())

    def test_get_parameters(self):
        cases = [
            ("<pyhp>print(pyhp.get['foo'])</pyhp>", {'foo': 'bar'}, 'bar\n'),
            ("<pyhp>print(pyhp.get['foo'] + ' ' + pyhp.get['baz'])</pyhp>",
             {'foo': 'bar', 'baz': 'qux'}, 'bar qux\n'),
        ]

        for case in cases:
            file_processor = MockFileProcessor()
            pyhp_class = Pyhp(PurePath(), file_processor, get=case[1])
            self.assertEqual(run_parsed_code(parse_text(case[0]), pyhp_class),
                             case[2])

    def test_post_parameters(self):
        cases = [
            ("<pyhp>print(pyhp.post['foo'])</pyhp>", {'foo': 'bar'}, 'bar\n'),
            ("<pyhp>print(pyhp.post['foo'] + ' ' + pyhp.post['baz'])</pyhp>",
             {'foo': 'bar', 'baz': 'qux'}, 'bar qux\n'),
        ]

        for case in cases:
            file_processor = MockFileProcessor()
            pyhp_class = Pyhp(PurePath(), file_processor, post=case[1])
            self.assertEqual(run_parsed_code(parse_text(case[0]), pyhp_class),
                             case[2])

    def test_include(self):
        pass  # TODO: Finish

    def test_display(self):
        file_processor = MockFileProcessor({
            PurePath('foo.html'): '<p>Hello</p>',
            PurePath('bar.pyhp'): '<pyhp>x=1\nprint(x + 2)</pyhp>',
        })
        pyhp_class = Pyhp(PurePath(), file_processor)

        self.assertEqual(
            run_parsed_code(parse_text('<pyhp>pyhp.display("foo.html")</pyhp>'),
                            pyhp_class),
            '<p>Hello</p>'
        )
        self.assertEqual(
            run_parsed_code(parse_text('<pyhp>pyhp.display("bar.pyhp")</pyhp>'),
                            pyhp_class),
            '3\n'
        )

    def test_run_error_code(self):
        cases = [
            ('<pyhp>x = </pyhp>', 'SyntaxError: invalid syntax'),
            ('<pyhp>raise Exception()</pyhp>', 'Exception'),
            ('<pyhp>print(x)</pyhp>', "NameError: name 'x' is not defined"),
        ]

        for case in cases:
            file_processor = MockFileProcessor()
            dom = parse_text(case[0])

            # Run with debug
            pyhp_class = Pyhp(PurePath(), file_processor, debug=True)
            self.assertIn(case[1], run_parsed_code(dom, pyhp_class))

            # Run without debug
            pyhp_class = Pyhp(PurePath(), file_processor, debug=False)
            with self.assertRaises(RuntimeError):
                run_parsed_code(dom, pyhp_class)

    def test_escape(self):
        cases = [
            ('<', '&lt;'),
            ('>', '&gt;'),
            ('&', '&amp;'),
            ('"', '&#34;'),
            ("'", '&#39;'),
            ('<script>', '&lt;script&gt;'),
            ('</script>', '&lt;/script&gt;'),
            ('<pyhp>print("Hello World!")</pyhp>',
             '&lt;pyhp&gt;print(&#34;Hello World!&#34;)&lt;/pyhp&gt;'),
        ]

        for case in cases:
            self.assertEqual(case[1], Pyhp.escape(case[0]))

    def test_redirect(self):
        cases = [
            ('<pyhp>pyhp.redirect("https://www.google.com")</pyhp>',
             ('https://www.google.com', 302)),
            ('<pyhp>pyhp.redirect("http://www.google.com", 301)</pyhp>',
             ('http://www.google.com', 301)),
            ('<pyhp>pyhp.redirect("test", 308)</pyhp>',
             ('test', 308)),
            ('<pyhp></pyhp>', None),
        ]

        for case in cases:
            file_processor = MockFileProcessor()
            pyhp_class = Pyhp(PurePath(), file_processor)
            run_parsed_code(parse_text(case[0]), pyhp_class)
            self.assertEqual(pyhp_class.get_redirect_information(), case[1])

    def test_import(self):
        pass  # TODO: Test import


class TestPyhpFileProcessing(TestCase):
    """Tests that PyHP can load and execute files."""

    def test_get_directory(self):
        pass  # TODO: Finish

    def test_get_true_path(self):
        file_contents = {'index.pyhp': '<pyhp>print("Hello")</pyhp>'}
        directories = ['dir', 'foo', 'foo/bar']

        file_processor = MockFileProcessor(
            {PurePath(f): contents for f, contents in file_contents.items()},
            {PurePath(d) for d in directories}
        )

        normal_cases = [
            ('index', 'index.pyhp'),
            ('index.pyhp', 'index.pyhp'),
        ]
        error_cases = [
            ('', ValueError),
            ('index.pyhp.txt', FileNotFoundError),
            ('dir', IsADirectoryError),
            ('foo/bar', IsADirectoryError),
            ('foo/baz', FileNotFoundError),
        ]

        for case in normal_cases:
            self.assertEqual(file_processor.get_true_path(PurePath(case[0])),
                             PurePath(case[1]))

        for case in error_cases:
            with self.assertRaises(case[1]):
                file_processor.get_true_path(PurePath(case[0]))


def create_file_processor_and_run_code(case: str):
    file_processor = MockFileProcessor(case)
    pyhp_class = Pyhp(PurePath(), file_processor)
    return pyhp_class.include('index.pyhp')



"""
PyHP: Python Hypertext Preprocessor

This is a Flask app that can be used to run PyHP files.

Run the examples with `set FLASK_APP=pyhp_flask:create_app('./examples')` and
`flask run`.
"""

from typing import Optional
from pathlib import Path, PurePath

from flask import Flask, request, make_response, Response, redirect, \
    send_from_directory

try:
    from pyhp_interface import Pyhp
    from file_processing import SystemFileProcessor
    from cookies import NewCookie, DeleteCookie
except ImportError:
    from .pyhp_interface import Pyhp
    from .file_processing import SystemFileProcessor
    from .cookies import NewCookie, DeleteCookie


def create_app(base_dir: str) -> Flask:
    """Create a PyHP Flask app and return it."""
    base_dir = Path(base_dir).absolute()

    app = Flask(__name__)

    @app.route('/', defaults={'path': 'index'}, methods=['GET', 'POST'])
    @app.route('/<path:path>', methods=['GET', 'POST'])
    def catch_all(path: str) -> Response:
        if path.endswith('/'):
            return redirect('index')

        file_processor = SystemFileProcessor(base_dir)

        try:
            relative_path = file_processor.get_true_path(PurePath(path))
        except FileNotFoundError:
            if app.config['DEBUG']:
                return make_response(f'File not found: {path}', 404)
            return make_response('', 404)
        except IsADirectoryError:
            return redirect(f'{path}/')

        current_dir = relative_path.parent

        pyhp_class = Pyhp(current_dir, file_processor, app.config['DEBUG'],
                          dict(request.cookies), dict(request.args),
                          dict(request.form))

        return process_request(file_processor,
                               PurePath(relative_path.name),
                               pyhp_class)

    return app


def process_request(file_processor: SystemFileProcessor,
                    relative_path: PurePath, pyhp_class: Pyhp) -> Response:
    """Process a request."""

    if file_processor.is_pyhp_file(relative_path):
        page_text = pyhp_class.run(str(relative_path))
        status_code = 200  # TODO: Change, allow for custom status codes

        return redirect_or_create_response(
            page_text, status_code,
            pyhp_class.get_new_cookies(),
            pyhp_class.get_delete_cookies(),
            pyhp_class.get_redirect_information())
    return send_from_directory(
        file_processor.get_absolute_path(pyhp_class.current_dir),
        relative_path
    )


def redirect_or_create_response(page_text: str, status_code: int,
                                new_cookies: dict[str, NewCookie],
                                delete_cookies: dict[str, DeleteCookie],
                                redirect_information: Optional[
                                    tuple[str, int]]) -> Response:
    """Return a redirect response or create a new response."""
    if redirect_information is not None:
        url, status_code = redirect_information
        return redirect(url, status_code)

    return create_response(page_text, status_code, new_cookies, delete_cookies)


def create_response(page_text: str, status_code: int,
                    new_cookies: dict[str, NewCookie],
                    delete_cookies: dict[str, DeleteCookie]) -> Response:
    """Create a new response."""
    response = make_response(page_text, status_code)

    for cookie in new_cookies.values():
        response.set_cookie(**cookie.__dict__)

    for cookie in delete_cookies.values():
        response.delete_cookie(**cookie.__dict__)

    return response



"""
Tests PyHP when connected to a flask server.

TestFileSecurity:
    Tests that the file security is correctly enforced.
"""

# pylint: disable=missing-function-docstring

from unittest import TestCase


class TestFileSecurity(TestCase):
    """Tests that the file security is correctly enforced."""
    def test_disallow_parent_directory_access(self):
        pass  # TODO: Finish

    def test_disallow_absolute_path_access(self):
        pass  # TODO: Finish



"""
Sets up functions for executing PyHP code blocks and code text.
"""

# pylint: disable=missing-function-docstring

from io import StringIO
from contextlib import redirect_stdout
from copy import deepcopy
from traceback import format_exc
from typing import TYPE_CHECKING, Any

from bs4 import BeautifulSoup, Tag

try:
    from text_processing import prepare_code_block
    from hypertext_processing import get_code_blocks
except ImportError:
    from .text_processing import prepare_code_block
    from .hypertext_processing import get_code_blocks

if TYPE_CHECKING:
    from .pyhp_interface import Pyhp


def run_parsed_code(dom: BeautifulSoup,
                    pyhp_class: 'Pyhp') -> str:
    output_dom = deepcopy(dom)
    code_blocks = get_code_blocks(output_dom)

    for code_block in code_blocks:
        success = run_code_block(code_block, pyhp_class.globals,
                                 pyhp_class.locals, pyhp_class)

        if not success:
            break

    return str(output_dom)


def run_code_block(code_block: Tag, globals_: dict[str, Any],
                   locals_: dict[str, Any], pyhp_class: 'Pyhp') -> bool:
    code_text = prepare_code_block(code_block)

    success, output = run_code_text(code_text, globals_, locals_)

    if not success and not pyhp_class.debug:
        raise RuntimeError(output)

    code_block.replace_with(BeautifulSoup(output, 'html.parser'))

    return success


def run_code_text(code_text: str,
                  globals_: dict[str, Any],
                  locals_: dict[str, Any]) -> (bool, str):
    output_text = StringIO()

    try:
        with redirect_stdout(output_text):
            exec(code_text, globals_, locals_)  # pylint: disable=exec-used
    except Exception:  # pylint: disable=broad-except
        success = False
        output = f'{output_text.getvalue()}<pre>{format_exc()}</pre>'
    else:
        success = True
        output = output_text.getvalue()

    return success, output



"""
Sets up classes for storing information about cookies that will be created and
destroyed.
"""

# pylint: disable=too-many-instance-attributes

from typing import Optional, Union
from dataclasses import dataclass
from datetime import datetime, timedelta


@dataclass
class NewCookie:
    """Stores information about a cookie that will be created."""
    key: str
    value: str = ''
    max_age: Optional[Union[timedelta, int]] = None
    expires: Optional[Union[str, datetime, int, float]] = None
    path: Optional[str] = '/'
    domain: Optional[str] = None
    secure: bool = False
    httponly: bool = False
    samesite: Optional[str] = None


@dataclass
class DeleteCookie:
    """Stores information about a cookie that will be destroyed."""
    key: str
    path: str = '/'
    domain: Optional[str] = None
    secure: bool = False
    httponly: bool = False
    samesite: Optional[str] = None



"""
Sets up functions for processing PyHP code blocks and code text.
"""

# pylint: disable=missing-function-docstring

from bs4 import Tag


def prepare_code_block(code_block: Tag) -> str:
    code_text = code_block.decode_contents()
    prepared_code_text = prepare_code_text(code_text)

    return prepared_code_text


def prepare_code_text(code_text: str) -> str:
    code_text = code_text.lstrip('\n').rstrip('\n')
    code_text = remove_initial_indentation(code_text)

    return code_text


def remove_initial_indentation(code_text: str) -> str:
    lines = code_text.split('\n')
    whitespace_characters = len(lines[0]) - len(lines[0].lstrip())

    new_lines = remove_whitespace_from_lines(lines, whitespace_characters)

    return '\n'.join(new_lines)


def remove_whitespace_from_lines(lines: list[str],
                                 whitespace_characters: int) -> list[str]:
    new_lines = []
    for line in lines:
        if can_remove_whitespace_from_line(line, whitespace_characters):
            new_lines.append(line[whitespace_characters:])
        else:
            raise SyntaxError('Invalid code block indentation. More '
                              'indentation was in the first line than in '
                              'subsequent lines.')

    return new_lines


def can_remove_whitespace_from_line(line: str,
                                    whitespace_characters: int) -> bool:
    return (line[:whitespace_characters].isspace() or
            whitespace_characters == 0 or
            line == '')



"""
Sets up functions for loading and parsing PyHP files.
"""

# pylint: disable=missing-function-docstring

from pathlib import PurePath, Path


PYHP_FILE_EXTENSION = 'pyhp'


class FileProcessor:
    """
    Abstract class for interacting with the file system.
    """
    def get_file_contents(self, path: PurePath) -> str:
        raise NotImplementedError

    def is_dir(self, path: PurePath) -> bool:
        raise NotImplementedError

    def is_file(self, path: PurePath) -> bool:
        raise NotImplementedError

    def is_pyhp_file(self, path: PurePath) -> bool:
        return path.suffix == f'.{PYHP_FILE_EXTENSION}'

    def get_true_path(self, path: PurePath) -> PurePath:
        if self.is_dir(path):
            raise IsADirectoryError

        if self.is_file(path):
            return path

        path = path.with_suffix(f'{path.suffix}.{PYHP_FILE_EXTENSION}')

        if self.is_file(path):
            return path

        raise FileNotFoundError('Unable to find file.')

    def get_absolute_path(self, path: PurePath) -> Path:
        raise NotImplementedError  # TODO: This is less than ideal


class SystemFileProcessor(FileProcessor):
    """
    Implementation of the FileProcessor interface, for normal interation with
    system files and directories.
    """
    def __init__(self, base_dir: Path):
        if not base_dir.is_dir():
            raise FileNotFoundError('The base directory does not exist.')
        self._base_dir = base_dir

    def get_file_contents(self, path: PurePath) -> str:
        with open(self.get_absolute_path(path), 'r', encoding='utf-8') as file:
            return file.read()

    def is_dir(self, path: PurePath) -> bool:
        return self.get_absolute_path(path).is_dir()

    def is_file(self, path: PurePath) -> bool:
        return self.get_absolute_path(path).is_file()

    def get_absolute_path(self, path: PurePath) -> Path:
        absolute_path = (self._base_dir / path).resolve()
        if self._base_dir not in absolute_path.parents and \
                self._base_dir != absolute_path:
            raise RuntimeError(f'Path ({absolute_path}) is outside of base '
                               f'directory ({self._base_dir}).')
        return absolute_path



"""
Interface for the PyHP programs.
"""

import sys
from typing import Optional, Any
from pathlib import PurePath
from bs4 import BeautifulSoup
import markupsafe

try:
    from file_processing import FileProcessor
    from code_execution import run_parsed_code
    from cookies import NewCookie, DeleteCookie
    from hypertext_processing import parse_text
except ImportError:
    from .file_processing import FileProcessor
    from .code_execution import run_parsed_code
    from .cookies import NewCookie, DeleteCookie
    from .hypertext_processing import parse_text

__all__ = ['Pyhp']


class Pyhp:
    """
    The interface for the PyHP file to interact with the server,
    and the web page.
    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments

    def __init__(self, current_dir: PurePath,
                 file_processor: FileProcessor,
                 debug: bool = False,
                 cookies: Optional[dict[str, str]] = None,
                 get: Optional[dict[str, str]] = None,
                 post: Optional[dict[str, str]] = None):
        self._current_dir = current_dir
        self._debug = debug
        self._file_processor = file_processor

        self._cookies: dict[str, str] = cookies or {}
        self._get: dict[str, str] = get or {}
        self._post: dict[str, str] = post or {}

        self._new_cookies: dict[str, NewCookie] = {}
        self._to_delete_cookies: dict[str, DeleteCookie] = {}

        self._redirect_info: Optional[(str, int)] = None

        self.globals, self.locals = self._prepare_context()

    def display(self, relative_path: str):
        """Include another pyhp file and print it."""
        print(self.include(relative_path), end='')

    def include(self, relative_path: str) -> str:
        """
        Include another pyhp file into the current one,
        and return the output HTML.
        """
        new_current_dir = (self._current_dir / PurePath(relative_path)).parent

        new_pyhp_class = Pyhp(new_current_dir, self._file_processor,
                              self._debug, self._cookies, self._get, self._post)

        return new_pyhp_class.run(PurePath(relative_path).name)

    def run(self, relative_path: str) -> str:
        """
        Runs another pyhp file in the context of the current one,
        and return the output HTML.
        """
        return run_parsed_code(
            self._parse_file(PurePath(relative_path)),
            self,
        )

    def _prepare_context(self) -> (dict[str, Any], dict[str, Any]):
        context_globals = {'pyhp': self}
        context_locals = {}

        # TODO: This is potentially a security flaw
        current_dir = str(
            self._file_processor.get_absolute_path(self.current_dir))

        if current_dir not in sys.path:
            sys.path.append(current_dir)

        return context_globals, context_locals

    def _parse_file(self, relative_path: PurePath) -> BeautifulSoup:
        return parse_text(self._load_file(relative_path))

    def _load_file(self, relative_path: PurePath) -> str:
        path = self._current_dir / relative_path
        return self._file_processor.get_file_contents(path)

    def redirect(self, url: str, status_code: int = 302):
        """Redirect to another url."""
        self._redirect_info = (url, status_code)

    def get_redirect_information(self) -> Optional[tuple[str, int]]:
        """Return the redirect information if present."""
        return self._redirect_info

    def set_cookie(self, key: str, **kwargs):
        """Set a cookie."""
        self._new_cookies[key] = NewCookie(key, **kwargs)

    def delete_cookie(self, key: str, **kwargs):
        """Delete a cookie."""
        self._to_delete_cookies[key] = DeleteCookie(key, **kwargs)

    def get_new_cookies(self):
        """Return the cookies that will be set."""
        return self._new_cookies

    def get_delete_cookies(self):
        """Return the cookies that will be deleted."""
        return self._to_delete_cookies

    @staticmethod
    def escape(text: str) -> str:
        """Escape text for use in HTML."""
        return str(markupsafe.escape(text))

    @property
    def current_dir(self) -> PurePath:
        """Return the directory of the currently executing file."""
        return self._current_dir

    @property
    def debug(self) -> bool:
        """Return whether the app is in debug mode."""
        return self._debug

    @property
    def cookies(self) -> dict[str, str]:
        """Return the cookies."""
        return self._cookies

    @property
    def get(self) -> dict[str, str]:
        """Return the GET parameters."""
        return self._get

    @property
    def post(self) -> dict[str, str]:
        """Return the POST data."""
        return self._post



"""
Sets up functions for processing DOMs and hypertext.
"""

# pylint: disable=missing-function-docstring

from bs4 import BeautifulSoup, Tag
from bs4.element import ResultSet

PYHP_TAG = 'pyhp'


def get_code_blocks(dom: BeautifulSoup) -> ResultSet[Tag]:
    return dom.select(f'{PYHP_TAG}:not({PYHP_TAG} *)')


def parse_text(text: str) -> BeautifulSoup:
    return BeautifulSoup(text, 'html.parser')
